!  sgssimul.f95
!  
!  Copyright 2015 Igor J. Chybicki <igorchy@ukw.edu.pl>
!  
!  This program is free software; you can redistribute it and/or modify
!  it under the terms of the GNU General Public License as published by
!  the Free Software Foundation; either version 2 of the License, or
!  (at your option) any later version.
!  
!  This program is distributed in the hope that it will be useful,
!  but WITHOUT ANY WARRANTY; without even the implied warranty of
!  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!  GNU General Public License for more details.
!  
!  You should have received a copy of the GNU General Public License
!  along with this program; if not, write to the Free Software
!  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
!  MA 02110-1301, USA.
!  
!  The code has been successfully compiled using GCC compiler 5.2.0.
!  In order to enable parallel routines the flag -fopenmp 
!  needs to be added while compiling. Some additional flags can 
!  further improve the preformance (e.g. -O3 -funroll-loops).
!  
!  The program relies on the initializing text file, which
!  must be written in advance. The file must be obligatory
!  named "info.txt". See read_info procedure for the details
!  regarding the content of the info file.

program sgssimul
   use omp_lib
   
   implicit none
   
   type individual
     integer, allocatable :: nugen(:,:)   !nuDNA genotype
     integer :: mthap,cphap               !mt and cpDNA haplotype
     double precision :: x,y,female       !coordinates, sex (prob. of femaleness)
     integer, dimension(2) :: sloc        !nuclear s-locus
   end type
   
   type(individual), allocatable :: gener(:,:) ! generated data (two generations)
   
   integer :: usernth,seed,ng,g,ncint,nrep,r,outfile1,outfile2
   real(8) :: tim
   character(len=99) :: fs,fc                ! output file names (slopes,correlograms)
   ! number of...
   integer :: ni,nl,na,nmt,ncp,nsl           ! individuals,nu loci,n alleles at nu,mt,cp and sloc
   double precision :: munu,mumt,mucp        ! mutation rates
   real(8) :: popdens                        ! population density [number of individuals / ha]
   real(8) :: femaleness                     ! proportion of females (0,1) (valid for dioecy=1)
   integer :: dioecy                         ! dioecy vs. monoecy indicator
   integer :: self_incomp                    ! self-incompatibility system enabled/disabled
   real(8) :: maternal_mt                    ! the probability of maternal transmitions of mtDNA
   real(8) :: maternal_cp                    ! the probability of maternal transmitions of cpDNA
   double precision :: a_s,b_s               ! exponential-power seed dispersal kernel
   double precision :: a_p,b_p               ! exponential-power pollen dispersal kernel
   double precision :: sig_mo,sig_fo,sig_mf  ! sigma for mother-offspring, father-offspring
   double precision :: dlt_mo,dlt_fo,dlt_mf  ! delta for mother-offspring, father-offspring
   double precision, allocatable :: phi(:,:),ksi(:,:),sphi(:),sksi(:)   ! repr. success rates
   double precision, allocatable :: pnu(:,:), pmt(:), pcp(:), psl(:)    ! allele frequencies at nu,mt,cp
   double precision, allocatable :: rnu(:,:), rmt(:,:), rcp(:,:),rsl(:,:) ! relatedness matrix for u,mt,cp
   integer, allocatable :: wij(:,:)          ! adjacency matrix for a series of distance classes
   double precision, allocatable :: uplim(:) ! upper limits for a series of distance classes
   
   ! correlogram (all arrays contain values per distance class)
   ! nu - nuclear, mt - mitochondrial, cp - chloroplast, sl - self-incompatibility nuclear locus
   integer, allocatable :: cnnu(:),cnmt(:),cncp(:),cnsl(:)          ! number of pairs
   double precision, allocatable :: crnu(:),crmt(:),crcp(:),crsl(:) ! mean relatedness
   double precision, allocatable :: cdnu(:),cdmt(:),cdcp(:),cdsl(:) ! meand distance
   double precision :: sl_nu,sl_mt,sl_cp,sl_sl              ! slopes of correlograms
   
   !RNG-related variables
   integer :: six,siy,siz,sit                            !serial version   
   integer, allocatable :: pix(:),piy(:),piz(:),pit(:)   !parallel version
   
 !---------------------- Main program starts here ----------------------
   
   outfile1=666
   outfile2=777
   
   call read_info
   
   call whrandomize(seed)
   call pwhrandomize(seed)
   
   open (unit=outfile1,file=fc,action="write",status="replace")
   open (unit=outfile2,file=fs,action="write",status="replace")
   
   tim = omp_get_wtime()
   
   do r=1,nrep
     call initialize_pop(ni,nl,na,nmt,ncp,nsl)
     call initialize_correlogram(ncint)
     write (*,'((a),(i0),(a))',advance='no') 'rep:',r,' '
     do g=1,ng
       call generate_progeny(ni)
       call generation_succession(ni)
       write (*,'((i0),(a))',advance='no') g,' '
     enddo
     write (*,*)
     
     call compute_correlogram(2,ni,nl,na,nmt,ncp,nsl,ncint)
     
     do g=1,ncint
       write (outfile1,'(2(i4.1),*(f10.4))') r,g,uplim(g),cdnu(g),crnu(g),crmt(g),crcp(g),crsl(g)
     enddo
     write (outfile2,'(1(i4.1),*(f10.4))') r,sl_nu,sl_mt,sl_cp,sl_sl,dlt_mo,dlt_fo,dlt_mf, &
      dsqrt(sig_mo/2.0),dsqrt(sig_fo/2.0),dsqrt(sig_mf/2.0)
     
     call flush(outfile1)
     call flush(outfile2)
   enddo
   
   close(outfile1)
   close(outfile2)
   
   tim = omp_get_wtime() - tim
   
   write (*,*) tim
   
   !call save_genotypes(2,ni,nl)
   
 !----------------------- Main program ends here -----------------------
 
 contains
 !procedure sorts 1-dimentional array of real values
 subroutine sort(ra,n)
   integer, intent(in) :: n
   double precision, intent(inout) :: ra(n)
   integer :: l,j,ir,i
   double precision :: rra
 
   l = (n / 2) + 1
   ir = n

   do
     if (l > 1) then
       l = l - 1
       rra = ra(l)
     else
       rra = ra(ir)
       ra(ir) = ra(1)
       ir = ir - 1
       if (ir == 1) then
         ra(1) = rra
         exit
       endif
     endif
   
     i = l
     j = 2 * l
     do while (j <= ir)
       if (j < ir) then
         if (ra(j) < ra(j+1)) then
           j = j + 1
         endif
       endif
     
       if (rra < ra(j)) then
         ra(i) = ra(j)
         i = j
         j = 2 * j
       else
         j = ir + 1 
       endif
     enddo
     ra(i) = rra
   enddo
 end subroutine sort
 !----------------------------------------------------------------------

 !routine initializes wh-RNG
 subroutine whrandomize(seed)
   integer, intent(in) :: seed
 
   call srand(seed)
   six=int(30000 * rand(0)) + 1
   siy=int(30000 * rand(0)) + 1
   siz=int(30000 * rand(0)) + 1
   sit=int(30000 * rand(0)) + 1
   
 end subroutine whrandomize
 !----------------------------------------------------------------------
 
 !function returns (0,1) random deviate
 double precision function whrnd()
   double precision :: v
   
   do
     
     six=11600 * mod(six,185127) - 10379 * (six / 185127)
     siy=47003 * mod(siy,45688) - 10479 * (siy / 45688)
     siz=23000 * mod(siz,93368) - 19423 * (siz / 93368)
     siz=33000 * mod(sit,65075) - 8123 * (sit / 65075)
     if (six < 0) then
       six= six + 2147483579
     endif
     if (siy < 0) then
       siy = siy + 2147483543
     endif
     if (siz < 0) then
       siz = siz + 2147483423
     endif
     if (sit < 0) then
       sit = sit + 2147483123
     endif

     v = (dble(six) / 2147483579) + (dble(siy) / 2147483543) &
       + (dble(siz) / 2147483423) + (dble(sit) / 2147483123)
     v = v - int(v)
     
     if (v > dble(1)) then
       v = v - dble(1)
     endif
    
     if (v.ne.dble(1)) then
       exit
     endif
     
   enddo

   whrnd=v
 end function whrnd
 !----------------------------------------------------------------------
 
 !function returns integer random deviate
 integer function whrndint(n)
   integer, intent(in) :: n
 
   whrndint = int(whrnd() * dble(n))
 end function whrndint
 !----------------------------------------------------------------------

 !routine initializes wh-RNG (parallel version)
 subroutine pwhrandomize(seed)
   integer, intent(in) :: seed
   integer :: i
   
   call srand(seed)
   
   allocate(pix(usernth))
   allocate(piy(usernth))
   allocate(piz(usernth))
   allocate(pit(usernth))
   
   do i=1,usernth
     pix(i)=int(30000 * rand(0)) + 1
     piy(i)=int(30000 * rand(0)) + 1
     piz(i)=int(30000 * rand(0)) + 1
     pit(i)=int(30000 * rand(0)) + 1
   enddo
 end subroutine pwhrandomize
 !----------------------------------------------------------------------
 
 !function returns (0,1) random deviate (parallel version)
 double precision function pwhrnd(ix,iy,iz,it)
   integer, intent(inout) :: ix,iy,iz,it
   double precision :: v
   
   do
     
     ix=11600 * mod(ix,185127) - 10379 * (ix / 185127)
     iy=47003 * mod(iy,45688) - 10479 * (iy / 45688)
     iz=23000 * mod(iz,93368) - 19423 * (iz / 93368)
     iz=33000 * mod(it,65075) - 8123 * (it / 65075)
     if (ix < 0) then
       ix= ix + 2147483579
     endif
     if (iy < 0) then
       iy = iy + 2147483543
     endif
     if (iz < 0) then
       iz = iz + 2147483423
     endif
     if (it < 0) then
       it = it + 2147483123
     endif

     v = (dble(ix) / 2147483579) + (dble(iy) / 2147483543) &
       + (dble(iz) / 2147483423) + (dble(it) / 2147483123)
     v = v - int(v)
     
     if (v > dble(1)) then
       v = v - dble(1)
     endif
    
     if (v.ne.dble(1)) then
       exit
     endif
     
   enddo

   pwhrnd=v
 end function pwhrnd
 !----------------------------------------------------------------------
 
 !function returns integer random deviate (parallel version)
 integer function pwhrndint(n,ix,iy,iz,it)
   integer, intent(in) :: n
   integer, intent(inout) :: ix,iy,iz,it
   
   pwhrndint = int(pwhrnd(ix,iy,iz,it) * dble(n))
 end function pwhrndint
 !----------------------------------------------------------------------
 
 !function generates random coordinates within a square
 subroutine generate_coordinates(g,n)
   integer, intent(in) :: g,n
   integer :: i,th,tix,tiy,tiz,tit
   double precision :: b

   b = dsqrt(dble(n) / dble(popdens) * dble(10000))
   
   !$omp parallel num_threads(usernth) private(tix,tiy,tiz,tit,th,i) &
   !$omp shared(pix,piy,piz,pit)
   !$omp do 
   do i=1,n
     th=omp_get_thread_num()+1
     
     tix=pix(th)
     tiy=piy(th)
     tiz=piz(th)
     tit=pit(th)
     
     gener(g,i)%x = pwhrnd(tix,tiy,tiz,tit) * b - b / dble(2)
     gener(g,i)%y = pwhrnd(tix,tiy,tiz,tit) * b - b / dble(2)
     
     pix(th)=tix
     piy(th)=tiy
     piz(th)=tiz
     pit(th)=tit
     
   enddo
   !$omp end do
   !$omp end parallel
   
 end subroutine generate_coordinates
 !----------------------------------------------------------------------
 
 !procedure initializes basic attributes of individuals
 subroutine initialize_genotypes_sex(g,n,nl,na,nmt,ncp,nsl)
   integer, intent(in) ::g,n,nl,na,nmt,ncp,nsl
   integer ::i,j,a
   double precision :: r

   do i=1,n
     if (.not. allocated(gener(g,i)%nugen)) allocate(gener(g,i)%nugen(nl,2))
     do j=1,nl
       gener(g,i)%nugen(j,1) = whrndint(na)+1
       gener(g,i)%nugen(j,2) = whrndint(na)+1
     enddo
     gener(g,i)%mthap = whrndint(nmt)+1
     gener(g,i)%cphap = whrndint(ncp)+1
     gener(g,i)%sloc(1) = whrndint(nsl)+1
     
     a = gener(g,i)%sloc(1)
     do while (a == gener(g,i)%sloc(1))
       a = whrndint(nsl)+1
     enddo
     gener(g,i)%sloc(2) = a
     
     if (dioecy == 1) then
       r = whrnd()
       if (r < femaleness) then
         gener(g,i)%female = dble(1)
       else
         gener(g,i)%female = dble(0)
       endif
     else
       gener(g,i)%female = dble(0.5)
     endif
   enddo
 end subroutine initialize_genotypes_sex
 !----------------------------------------------------------------------
 
 !procedure initializes some variables
 subroutine initialize_pop(ni,nl,na,nmt,ncp,nsl)
   integer, intent(in) :: ni,nl,na,nmt,ncp,nsl
   
   if (.not. allocated(pnu)) allocate(pnu(nl,na))
   if (.not. allocated(pmt)) allocate(pmt(nmt))
   if (.not. allocated(pcp)) allocate(pcp(ncp))
   if (.not. allocated(psl)) allocate(psl(nsl))
   
   if (.not. allocated(rnu)) allocate(rnu(ni,ni))
   if (.not. allocated(rmt)) allocate(rmt(ni,ni))
   if (.not. allocated(rcp)) allocate(rcp(ni,ni))
   if (.not. allocated(rsl)) allocate(rsl(ni,ni))
   
   if (.not. allocated(wij)) allocate(wij(ni,ni));
   
   if (.not. allocated(gener)) allocate(gener(2,ni))
   if (.not. allocated(ksi)) allocate(ksi(ni,ni))
   if (.not. allocated(phi)) allocate(phi(ni,ni))
   if (.not. allocated(sksi)) allocate(sksi(ni))
   if (.not. allocated(sphi)) allocate(sphi(ni))
   
   call generate_coordinates(1,ni)
   call initialize_genotypes_sex(1,ni,nl,na,nmt,ncp,nsl)
   call initialize_genotypes_sex(2,ni,nl,na,nmt,ncp,nsl) 
   
   sig_mo = 0.0
   sig_fo = 0.0
   sig_mf = 0.0
   dlt_mo = 0.0
   dlt_fo = 0.0
   dlt_mf = 0.0
 end subroutine initialize_pop
 !----------------------------------------------------------------------
 
 double precision function compatibility(m1,m2,f1,f2)
   integer, intent(in) :: m1,m2,f1,f2
   integer :: p
   
   p=0
   if (m1/=f1) then
     p=p+1
   end if
   
  if (m1/=f2) then
     p=p+1
   end if
   
  if (m2/=f1) then
     p=p+1
   end if
   
  if (m2/=f2) then
     p=p+1
   end if
   
   compatibility=dble(self_incomp) * dble(p/4) + dble(1-self_incomp)
 end function compatibility
 
 !procedure computes female reproductive success of the i-th female
 subroutine compute_ksi(i,ni)
   integer, intent(in) :: i,ni
   integer :: j
   double precision :: d
   
   sksi(i) = dble(0)
   do j=1,ni
     if (gener(1,j)%female > dble(0)) then
       d = dsqrt((gener(1,j)%x-gener(2,i)%x)**2 + (gener(1,j)%y-gener(2,i)%y)**2)
       ksi(j,i) = dexp(-((d/a_s)**b_s))
       sksi(i) = sksi(i) + ksi(j,i)
     else
       ksi(j,i) = dble(0)
     endif  
   enddo
 end subroutine compute_ksi
 !----------------------------------------------------------------------
 
 !procedure computes male reproductive success of the i-th male
 subroutine compute_phi(i,ni)
   integer, intent(in) :: i,ni
   integer :: j,m1,m2,f1,f2
   double precision :: d
   
   m1=gener(1,i)%sloc(1)
   m2=gener(1,i)%sloc(2)
   sphi(i) = dble(0)
   do j=1,ni
     if (gener(1,j)%female < dble(1)) then
       d = dsqrt((gener(1,j)%x-gener(1,i)%x)**2 + (gener(1,j)%y-gener(1,i)%y)**2)
       f1=gener(1,j)%sloc(1)
       f2=gener(1,j)%sloc(2)
       phi(j,i) = compatibility(m1,m2,f1,f2) * dexp(-((d/a_p)**b_p))
       sphi(i) = sphi(i) + phi(j,i)
     else
       phi(j,i) = dble(0)
     endif
   enddo
 end subroutine compute_phi
 !----------------------------------------------------------------------
 
 !procedure computes reproductive success rates
 subroutine reproductive_success(ni)
   integer, intent(in) :: ni
   integer :: i,nth
   
   nth=usernth
   
   !$omp parallel num_threads(nth) private(i) &
   !$omp shared(ksi,sksi,phi,sphi,ni)
   !$omp do
   do i=1,ni
     call compute_ksi(i,ni)
     call compute_phi(i,ni)
   enddo
   !$omp end do
   !$omp end parallel
   
 end subroutine reproductive_success
 !----------------------------------------------------------------------
 
 integer function ssr_mutant(a,n,ix,iy,iz,it)
   integer, intent(inout) :: ix,iy,iz,it
   integer, intent(in) :: a,n
   double precision :: r
   integer :: ma
   
   r=pwhrnd(ix,iy,iz,it)
   if (r < 0.5) then
     if (a+1>n) then
       ma=1
     else
       ma=a+1
     endif
   else
     if (a-1 < 1) then
       ma=n
     else
       ma=a-1
     endif
   endif
   
   ssr_mutant=ma
   
 end function ssr_mutant
 
 !procedure generates single progeny
 subroutine generate_sngl_progeny(i,ni,ix,iy,iz,it,d_mo,d_fo,d_mf,s_mo,s_fo,s_mf)
   integer, intent(in) :: i,ni
   integer, intent(inout) :: ix,iy,iz,it
   double precision, intent(inout) :: d_mo,d_fo,d_mf,s_mo,s_fo,s_mf
   integer :: m,f,j,a1,a2
   double precision :: r,s,d
   
   m = 1
   f = 1
   
   r = pwhrnd(ix,iy,iz,it)

   s = dble(0)
   do j=1,ni
     s = s + ksi(j,i) / sksi(i)
     if (r >= s - ksi(j,i) / sksi(i) .and. r < s) then
       m = j
       exit
     endif
   enddo

   r = pwhrnd(ix,iy,iz,it)

   s = dble(0)
   do j=1,ni
     s = s + phi(j,m) / sphi(m)
     if (r >= s - phi(j,m) / sphi(m) .and. r < s) then
       f = j
       exit
     endif
   enddo
   
   d = (gener(1,m)%x-gener(2,i)%x)**2 + (gener(1,m)%y-gener(2,i)%y)**2
   d_mo=dsqrt(d) / dble(ni) / dble(ng)
   s_mo=d / dble(ni) / dble(ng)
   
   d = (gener(1,f)%x-gener(2,i)%x)**2 + (gener(1,f)%y-gener(2,i)%y)**2
   d_fo=dsqrt(d) / dble(ni) / dble(ng)
   s_fo=d / dble(ni) / dble(ng)
   
   d = (gener(1,f)%x-gener(1,m)%x)**2 + (gener(1,f)%y-gener(1,m)%y)**2
   d_mf=dsqrt(d) / dble(ni) / dble(ng)
   s_mf=d / dble(ni) / dble(ng)
   
   do j=1,nl
     gener(2,i)%nugen(j,1) = gener(1,m)%nugen(j,pwhrndint(2,ix,iy,iz,it)+1)
     gener(2,i)%nugen(j,2) = gener(1,f)%nugen(j,pwhrndint(2,ix,iy,iz,it)+1)
     
     if (pwhrnd(ix,iy,iz,it) < munu) then
       gener(2,i)%nugen(j,1)=ssr_mutant(gener(2,i)%nugen(j,1),na,ix,iy,iz,it)
     endif
     
     if (pwhrnd(ix,iy,iz,it) < munu) then
       gener(2,i)%nugen(j,2)=ssr_mutant(gener(2,i)%nugen(j,2),na,ix,iy,iz,it)
     endif
   enddo
     
   if (pwhrnd(ix,iy,iz,it) < maternal_mt) then
     gener(2,i)%mthap = gener(1,m)%mthap
   else
     gener(2,i)%mthap = gener(1,f)%mthap
   endif
   
   if (pwhrnd(ix,iy,iz,it) < mumt) then
     gener(2,i)%mthap=ssr_mutant(gener(2,i)%mthap,nmt,ix,iy,iz,it)
   endif
   
   if (pwhrnd(ix,iy,iz,it) < maternal_cp) then
     gener(2,i)%cphap = gener(1,m)%cphap
   else
     gener(2,i)%cphap = gener(1,f)%cphap;
   endif
   
   if (pwhrnd(ix,iy,iz,it) < mucp) then
     gener(2,i)%cphap=ssr_mutant(gener(2,i)%cphap,ncp,ix,iy,iz,it)
   endif
   
   a1 = gener(1,m)%sloc(pwhrndint(2,ix,iy,iz,it)+1)
   a2 = gener(1,f)%sloc(pwhrndint(2,ix,iy,iz,it)+1)
   if (self_incomp == 1) then
     do while (a1 == a2)
     a1 = gener(1,m)%sloc(pwhrndint(2,ix,iy,iz,it)+1)
     a2 = gener(1,f)%sloc(pwhrndint(2,ix,iy,iz,it)+1)
     enddo
   endif
   gener(2,i)%sloc(1) = a1
   gener(2,i)%sloc(2) = a2
   
   if (dioecy == 1) then
     r = pwhrnd(ix,iy,iz,it)
     if (r < femaleness) then
       gener(2,i)%female = dble(1)
     else
       gener(2,i)%female = dble(0)
     endif
   else
     gener(2,i)%female = dble(0.5)
   endif
   
 end subroutine generate_sngl_progeny
 !----------------------------------------------------------------------
 
 !procedure generates progeny array
 subroutine generate_progeny(ni)
   integer, intent(in) :: ni
   integer :: i,tix,tiy,tiz,tit,th,nth
   double precision :: d_mo,d_fo,d_mf,s_mo,s_fo,s_mf
   
   nth=usernth
   
   call generate_coordinates(2,ni)
   call reproductive_success(ni)
   
   !$omp parallel num_threads(nth) private(tix,tiy,tiz,tit,th,i) &
   !$omp shared(pix,piy,piz,pit,dlt_mo,dlt_fo,dlt_mf,sig_mo,sig_fo,sig_mf)
   !$omp do reduction(+:dlt_mo,dlt_fo,dlt_mf,sig_mo,sig_fo,sig_mf)
   do i=1,ni
     th=omp_get_thread_num()+1
     
     tix=pix(th)
     tiy=piy(th)
     tiz=piz(th)
     tit=pit(th)
     
     call generate_sngl_progeny(i,ni,tix,tiy,tiz,tit,d_mo,d_fo,d_mf,s_mo,s_fo,s_mf)
     
     sig_mo=sig_mo+s_mo
     sig_fo=sig_fo+s_fo
     sig_mf=sig_mf+s_mf
     dlt_mo=dlt_mo+d_mo
     dlt_fo=dlt_fo+d_fo
     dlt_mf=dlt_mf+d_mf
     
     pix(th)=tix
     piy(th)=tiy
     piz(th)=tiz
     pit(th)=tit
   enddo
   !$omp end do
   !$omp end parallel
   
 end subroutine generate_progeny      
 !----------------------------------------------------------------------
 
 !procedure substitutes generations
 subroutine generation_succession(ni)
   integer, intent(in) :: ni
   integer :: i,nth
   
   nth=usernth
   
   !$omp parallel num_threads(nth) private(i) &
   !$omp shared(gener)
   !$omp do 
   do i=1,ni
     gener(1,i) = gener(2,i)
   enddo
   !$omp end do
   !$omp end parallel
   
 end subroutine generation_succession
 !----------------------------------------------------------------------
 
 !procedure computes allele frequencies
 subroutine compute_pij(g,ni,nl,na,nmt,ncp,nsl)
   integer, intent(in) :: g,ni,nl,na,nmt,ncp,nsl
   integer :: i,j
 
   do i=1,nl
     do j=1,na
       pnu(i,j) = dble(0)
     enddo
   enddo

   do i=1,nmt
     pmt(i) = dble(0)
   enddo

   do i=1,ncp
     pcp(i) = dble(0)
   enddo
   
   do i=1,nsl
     psl(i)= dble(0)
   enddo

   do i=1,ni
     do j=1,nl
       pnu(j,gener(g,i)%nugen(j,1)) = pnu(j,gener(g,i)%nugen(j,1)) + dble(0.5) / dble(ni)
       pnu(j,gener(g,i)%nugen(j,2)) = pnu(j,gener(g,i)%nugen(j,2)) + dble(0.5) / dble(ni)
     enddo
   pmt(gener(g,i)%mthap) = pmt(gener(g,i)%mthap) + dble(1) / dble(ni)
   pcp(gener(g,i)%cphap) = pcp(gener(g,i)%cphap) + dble(1) / dble(ni)
   
   psl(gener(g,i)%sloc(1)) = psl(gener(g,i)%sloc(1)) + dble(0.5) / dble(ni)
   psl(gener(g,i)%sloc(2)) = psl(gener(g,i)%sloc(2)) + dble(0.5) / dble(ni)
   enddo
 end subroutine compute_pij
 !---------------------------------------------------------------------- 
 
 !procedure computes upper limits for correlogram
 subroutine compute_uplimits(g,ni,nintervals)
   integer, intent(in) :: g,ni,nintervals
   integer :: i,j,k,n
   double precision, allocatable :: dij(:)
   double precision :: d
   
   n = ni * (ni - 1) / 2
   allocate(dij(n))

   k = 0
   do i=1,ni
     do j=i+1,ni
       d = dsqrt((gener(g,i)%x - gener(g,j)%x)**2 + (gener(g,i)%y - gener(g,j)%y)**2)

       k=k+1
       
       dij(k) = d
     enddo
   enddo
   
   call sort(dij,n)

   do i=1,nintervals
     uplim(i) = dij((i) * (k / nintervals))
   enddo
 end subroutine compute_uplimits
 !---------------------------------------------------------------------- 
 
 !procedure computes single value in adjacency matrix
 subroutine compute_sngl_wij(g,i,j,nintervals)
   integer, intent(in) :: g,i,j,nintervals
   integer :: k
   double precision :: d
   
   d = dsqrt((gener(g,i)%x - gener(g,j)%x)**2 + (gener(g,i)%y - gener(g,j)%y)**2)
   wij(i,j) = 0
   do k=1,nintervals
     if (d < uplim(k)) then
       wij(i,j) = k
       exit
     endif
   enddo
      
   wij(j,i) = wij(i,j)
 end subroutine compute_sngl_wij
 !----------------------------------------------------------------------
 
 !procedure computes adjacency matrix
 subroutine compute_wij(g,ni,nintervals)
   integer, intent(in) :: g,ni,nintervals
   integer :: i,j,nth
   
   nth=usernth
   
   !$omp parallel num_threads(nth) private(i,j)
   !$omp do 
   do i=1,ni
     do j=i+1,ni
       call compute_sngl_wij(g,i,j,nintervals)
     enddo
   enddo
   !$omp end do
   !$omp end parallel
   
 end subroutine compute_wij 
 !----------------------------------------------------------------------  
 
 !procedure computes single relatedness between i and j
 subroutine compute_sngl_relatedness(g,ni,nl,na,nmt,ncp,nsl,dnu,dmt,dcp,dsl,i,j)
   integer, intent(in) :: g,ni,nl,na,nmt,ncp,nsl,i,j
   double precision, intent(in) :: dnu,dmt,dcp,dsl
   double precision :: n,p_i,p_j
   integer :: k,l
   
   n = dble(0)
   do k=1,nl
     do l=1,na
       p_i = dble(0)
       p_j = dble(0)
       if (gener(g,i)%nugen(k,1) == l) then 
         p_i = dble(0.5)
       endif
       if (gener(g,i)%nugen(k,2) == l) then
         p_i = p_i + dble(0.5)
       endif
       if (gener(g,j)%nugen(k,1) == l) then
         p_j = dble(0.5)
       endif
       if (gener(g,j)%nugen(k,2) == l) then
         p_j = p_j + dble(0.5)
       endif

       n = n + (p_i - pnu(k,l)) * (p_j - pnu(k,l))
     enddo
   enddo
      
   if (dnu > 0) then
     rnu(i,j) = n / dnu + dble(1) / dble(ni - 1)
   else
     rnu(i,j) = dble(-9999)
   endif
   rnu(j,i) = rnu(i,j)

   n = dble(0)
   do k=1,nmt
     p_i = dble(0)
     p_j = dble(0)
     if (gener(g,i)%mthap == k) then
       p_i = dble(1)
     endif
     if (gener(g,j)%mthap == k) then
       p_j = dble(1)
     endif
     n = n + (p_i - pmt(k)) * (p_j - pmt(k))
   enddo
   
   if (dmt > 0) then
     rmt(i,j) = n / dmt + dble(1) / dble(ni - 1)
   else 
     rmt(i,j) = dble(-9999)
   endif
   rmt(j,i) = rmt(i,j)
   
   n = dble(0)
   do k=1,ncp
     p_i = dble(0)
     p_j = dble(0)
     if (gener(g,i)%cphap == k) then
       p_i = dble(1)
     endif
     if (gener(g,j)%cphap == k) then
       p_j = dble(1)
     endif
     n = n + (p_i - pcp(k)) * (p_j - pcp(k))
   enddo
   
   if (dcp > 0) then
     rcp(i,j) = n / dcp + dble(1) / dble(ni - 1)
   else 
     rcp(i,j) = dble(-9999)
   endif
   rcp(j,i) = rcp(i,j)
   
   n = dble(0)
   do l=1,nsl
     p_i = dble(0)
     p_j = dble(0)
     if (gener(g,i)%sloc(1) == l) then 
       p_i = dble(0.5)
     endif
     if (gener(g,i)%sloc(2) == l) then
       p_i = p_i + dble(0.5)
     endif
     if (gener(g,j)%sloc(1) == l) then
       p_j = dble(0.5)
     endif
     if (gener(g,j)%sloc(2) == l) then
       p_j = p_j + dble(0.5)
     endif
     n = n + (p_i - psl(l)) * (p_j - psl(l))
   enddo
      
   if (dsl > 0) then
     rsl(i,j) = n / dsl + dble(1) / dble(ni - 1)
   else
     rsl(i,j) = dble(-9999)
   endif
   rsl(j,i) = rsl(i,j)
 end subroutine compute_sngl_relatedness
 !----------------------------------------------------------------------
 
 !procedure computes relatedness matrices for nu, mt and cp
 subroutine compute_relatedness(g,ni,nl,na,nmt,ncp)
   integer, intent(in) :: g,ni,nl,na,nmt,ncp
   integer :: i,j,k,l,nth
   double precision :: p_i,dnu,dmt,dcp,dsl
   
   dnu = dble(0)
   dmt = dble(0)
   dcp = dble(0)
   dsl = dble(0)
   do i=1,ni
     do k=1,nl
       do l=1,na
         p_i = dble(0)
         if (gener(g,i)%nugen(k,1) == l) then
           p_i = dble(0.5)
         endif
         if (gener(g,i)%nugen(k,2) == l) then
           p_i = p_i + dble(0.5)
         endif
         dnu = dnu + (p_i - pnu(k,l))**2
       enddo
     enddo
     
     do l=1,nmt
       p_i = dble(0)
       if (gener(g,i)%mthap == l) then
         p_i = 1
       endif
     dmt = dmt + (p_i - pmt(l))**2
     enddo

     do l=1,ncp
       p_i = dble(0)
       if (gener(g,i)%cphap == l) then
         p_i = 1
       endif
     dcp = dcp + (p_i - pcp(l))**2
     enddo
     
     do l=1,nsl
       p_i = dble(0)
       if (gener(g,i)%sloc(1) == l) then
         p_i = dble(0.5)
       endif
       if (gener(g,i)%sloc(2) == l) then
         p_i = p_i + dble(0.5)
       endif
     dsl = dsl + (p_i - psl(l))**2
     enddo
   enddo
   
   
   dnu = dnu / dble(ni)
   dmt = dmt / dble(ni)
   dcp = dcp / dble(ni)
   dsl = dsl / dble(ni)

   nth=usernth
   
   !$omp parallel num_threads(nth) private(i,j)
   !$omp do 
   do i=1,ni
     do j=i+1,ni
       call compute_sngl_relatedness(g,ni,nl,na,nmt,ncp,nsl,dnu,dmt,dcp,dsl,i,j)
     enddo
   enddo
   !$omp end do
   !$omp end parallel
   
 end subroutine compute_relatedness
 !----------------------------------------------------------------------
 
 !procedure initializes correlograms
 subroutine initialize_correlogram(nintervals)
   integer, intent(in) :: nintervals
   
   if (.not. allocated(cnnu)) allocate(cnnu(nintervals))
   if (.not. allocated(cnmt)) allocate(cnmt(nintervals))
   if (.not. allocated(cncp)) allocate(cncp(nintervals))
   if (.not. allocated(cnsl)) allocate(cnsl(nintervals))
   if (.not. allocated(crnu)) allocate(crnu(nintervals))
   if (.not. allocated(crmt)) allocate(crmt(nintervals))
   if (.not. allocated(crcp)) allocate(crcp(nintervals))
   if (.not. allocated(crsl)) allocate(crsl(nintervals))
   if (.not. allocated(cdnu)) allocate(cdnu(nintervals))
   if (.not. allocated(cdmt)) allocate(cdmt(nintervals))
   if (.not. allocated(cdcp)) allocate(cdcp(nintervals))
   if (.not. allocated(cdsl)) allocate(cdsl(nintervals))
   if (.not. allocated(uplim)) allocate(uplim(nintervals))
 end subroutine initialize_correlogram
 !----------------------------------------------------------------------
 
 !procedure computes correlograms for nu, mt and cp
 subroutine compute_correlogram(g,ni,nl,na,nmt,ncp,nsl,nintervals)
   integer, intent(in) :: g,ni,nl,na,nmt,ncp,nsl,nintervals
   integer :: i,j,n_nu,n_mt,n_cp,n_sl
   double precision :: d,sfd_nu,sf_nu,sd_nu,sd2_nu
   double precision :: sfd_mt,sf_mt,sd_mt,sd2_mt
   double precision :: sfd_cp,sf_cp,sd_cp,sd2_cp
   double precision :: sfd_sl,sf_sl,sd_sl,sd2_sl
   
   call compute_pij(g,ni,nl,na,nmt,ncp,nsl)
   call compute_relatedness(g,ni,nl,na,nmt,ncp)
   call compute_uplimits(g,ni,nintervals)
   call compute_wij(g,ni,nintervals)
   
   do i=1,nintervals
     crnu(i) = dble(0)
     crmt(i) = dble(0)
     crcp(i) = dble(0)
     crsl(i) = dble(0)
     cnnu(i) = 0
     cnmt(i) = 0
     cncp(i) = 0
     cnsl(i) = 0
     cdnu(i) = dble(0)
     cdmt(i) = dble(0)
     cdcp(i) = dble(0)
     cdsl(i) = dble(0)
   enddo
   
   sfd_nu = dble(0)
   sf_nu = dble(0)
   sd_nu = dble(0)
   sd2_nu = dble(0)
   n_nu = 0
   
   sfd_mt = dble(0)
   sf_mt = dble(0)
   sd_mt = dble(0)
   sd2_mt = dble(0)
   n_mt = 0
   
   sfd_cp = dble(0)
   sf_cp = dble(0)
   sd_cp = dble(0)
   sd2_cp = dble(0)
   n_cp = 0
   
   sfd_sl = dble(0)
   sf_sl = dble(0)
   sd_sl = dble(0)
   sd2_sl = dble(0)
   n_sl = 0
   do i=1,ni
     do j=i + 1,ni
       d = sqrt((gener(g,i)%x - gener(g,j)%x)**2 + (gener(g,i)%y - gener(g,j)%y)**2)
       if (wij(i,j) > 0 .and. wij(i,j) <= nintervals .and. rnu(i,j) /= dble(-9999)) then
         crnu(wij(i,j)) = crnu(wij(i,j)) + rnu(i,j)
         cdnu(wij(i,j)) = cdnu(wij(i,j)) + d
         cnnu(wij(i,j)) = cnnu(wij(i,j)) + 1
         n_nu = n_nu + 1
         sf_nu = sf_nu + rnu(i,j)
         sd_nu = sd_nu + dlog(d)
         sfd_nu = sfd_nu + rnu(i,j) * dlog(d)
         sd2_nu = sd2_nu + (dlog(d))**2
       endif

       if (wij(i,j) > 0 .and. wij(i,j) <= nintervals .and. rmt(i,j) /= dble(-9999)) then
         crmt(wij(i,j)) = crmt(wij(i,j)) + rmt(i,j)
         cdmt(wij(i,j)) = cdmt(wij(i,j)) + d
         cnmt(wij(i,j)) = cnmt(wij(i,j)) + 1
         n_mt = n_mt + 1
         sf_mt = sf_mt + rmt(i,j)
         sd_mt = sd_mt + dlog(d)
         sfd_mt = sfd_mt + rmt(i,j) * dlog(d)
         sd2_mt = sd2_mt + (dlog(d))**2
       endif

       if (wij(i,j) > 0 .and. wij(i,j) <= nintervals .and. rcp(i,j) /= dble(-9999)) then
         crcp(wij(i,j)) = crcp(wij(i,j)) + rcp(i,j)
         cdcp(wij(i,j)) = cdcp(wij(i,j)) + d
         cncp(wij(i,j)) = cncp(wij(i,j)) + 1
         n_cp = n_cp + 1
         sf_cp = sf_cp + rcp(i,j)
         sd_cp = sd_cp + dlog(d)
         sfd_cp = sfd_cp + rcp(i,j) * dlog(d)
         sd2_cp = sd2_cp + (dlog(d))**2
       endif
       
       if (wij(i,j) > 0 .and. wij(i,j) <= nintervals .and. rsl(i,j) /= dble(-9999)) then
         crsl(wij(i,j)) = crsl(wij(i,j)) + rsl(i,j)
         cdsl(wij(i,j)) = cdsl(wij(i,j)) + d
         cnsl(wij(i,j)) = cnsl(wij(i,j)) + 1
         n_sl = n_sl + 1
         sf_sl = sf_sl + rsl(i,j)
         sd_sl = sd_sl + dlog(d)
         sfd_sl = sfd_sl + rsl(i,j) * dlog(d)
         sd2_sl = sd2_sl + (dlog(d))**2
       endif
     enddo
   enddo
   
   sl_nu = (dble(n_nu) * sfd_nu - sf_nu * sd_nu) / (dble(n_nu) * sd2_nu - sd_nu**2)
   sl_mt = (dble(n_mt) * sfd_mt - sf_mt * sd_mt) / (dble(n_mt) * sd2_mt - sd_mt**2)
   sl_cp = (dble(n_cp) * sfd_cp - sf_cp * sd_cp) / (dble(n_cp) * sd2_cp - sd_cp**2)
   sl_sl = (dble(n_sl) * sfd_sl - sf_sl * sd_sl) / (dble(n_sl) * sd2_sl - sd_sl**2)
   
   do i=1,nintervals
     if (cnnu(i) > dble(0)) then
       crnu(i) = crnu(i) / dble(cnnu(i))
       cdnu(i) = cdnu(i) / dble(cnnu(i))
     endif
     
     if (cnmt(i) > dble(0)) then
       crmt(i) = crmt(i) / dble(cnmt(i))
       cdmt(i) = cdmt(i) / dble(cnmt(i))
     endif
     
     if (cncp(i) > dble(0)) then
       crcp(i) = crcp(i) / dble(cncp(i))
       cdcp(i) = cdcp(i) / dble(cncp(i))
     endif
     
     if (cnsl(i) > dble(0)) then
       crsl(i) = crsl(i) / dble(cnsl(i))
       cdsl(i) = cdsl(i) / dble(cnsl(i))
     endif
   enddo

 end subroutine compute_correlogram
 !----------------------------------------------------------------------
 
 !procedure saves genotypes to text file
 subroutine save_genotypes(g,ni,nl)
   integer, intent(in) :: g,ni,nl
   integer :: i,j,k,out_unit
   
   out_unit = 20
   open (unit=out_unit,file='genotypes.txt',action="write",status="replace")
   do i=1,ni
     write(out_unit,"(*((i0),' '))") ((gener(g,i)%nugen(j,k),k=1,2),j=1,nl),gener(g,i)%mthap,gener(g,i)%cphap
   enddo
   close(out_unit)
 end subroutine save_genotypes
 !----------------------------------------------------------------------
 
 !procedure reads info file
 subroutine read_info
   integer :: infile
   
   infile=44
   
   open (unit=infile, file='info.txt')
   read(infile,*) nrep                 !number of replicates
   read(infile,*) ng                   !number of generations
   read(infile,*) ni                   !number of individuals
   read(infile,*) popdens              !population density
   read(infile,*) femaleness           !proportion of females
   read(infile,*) dioecy               !dioecy vs. monoecy indicator
   read(infile,*) self_incomp          !self-infcomp. indicator
   read(infile,*) a_s,b_s              !seed dispersal kernel
   read(infile,*) a_p,b_p              !pollen dispersal kernel
   read(infile,*) maternal_mt          !probability of maternal transmission of mtDNA
   read(infile,*) maternal_cp          !probability of maternal transmission of cpDNA
   read(infile,*) nl                   !number of nuclear loci
   read(infile,*) na                   !number of alleles at nuclear loci
   read(infile,*) nmt                  !number of mt haplotypes
   read(infile,*) ncp                  !number of cp haplotypes
   read(infile,*) nsl                  !number of s-locus alleles
   read(infile,*) munu,mumt,mucp       !mutation rates
   read(infile,*) ncint                !number of distance intervals to compute correlogram
   read(infile,*) seed                 !seed to initialize random number generator
   read(infile,*) fs                   !output file name (max=99 characters!)
   read(infile,*) fc                   !output file name (max=99 characters!)
   read(infile,*) usernth              !number of threads
   if (usernth > omp_get_max_threads()) then
     usernth = omp_get_max_threads()
   endif
   
   close(infile)
 end subroutine read_info
 !----------------------------------------------------------------------

end program sgssimul
